package com.intellij.indexing.shared.builder

import com.beust.jcommander.*
import java.nio.file.Path
import java.nio.file.Paths
import kotlin.io.path.absolutePathString


fun main(args: Array<String>) {
  val versionCommand = VersionCommand()
  val serverCommand = ServerCommand()
  val boostCommand = BoostCommand()
  val indexesCommand = IndexesCommand()

  val jc = JCommander
    .newBuilder()
    .addCommand(VERSION_CMD_NAME, versionCommand)
    .addCommand(SERVER_CMD_NAME, serverCommand)
    .addCommand(BOOST_CMD_NAME, boostCommand)
    .addCommand(INDEXES_CMD_NAME, indexesCommand)
    .build()

  if (args.isNotEmpty() && args[0].endsWith(".json")) {
    val jsonFilePath = args[0]
    processConfigFromJson(jsonFilePath)
    return
  }

  try {
    jc.parse(*args)
  } catch (e: ParameterException) {
    println(e.message)
    println()
    printHelp()
    return
  }

  when (jc.parsedCommand) {
    VERSION_CMD_NAME -> {
      buildSharedIndexesVersion(versionCommand.ij!!)
    }

    SERVER_CMD_NAME -> {
      startServerOnLocalIndexes(serverCommand.serverDir, serverCommand.port)
    }

    BOOST_CMD_NAME -> {
      measureSharedIndexesBoost(
        boostCommand.project!!,
        boostCommand.ij!!,
        boostCommand.dataDir,
        boostCommand.port
      )
    }

    INDEXES_CMD_NAME -> {
      buildSharedIndexes(
        indexesCommand.project!!,
        indexesCommand.ij!!,
        indexesCommand.baseUrl!!,
        indexesCommand.dataDir
      )
    }

    else -> {
      if (jc.parsedCommand != null) {
        println("Unknown command: ${jc.parsedCommand}")
        println()
      }
      printHelp()
    }
  }
}

private const val VERSION_CMD_NAME = "version"
private const val VERSION_CMD_DESCRIPTION = "Displays shared indexes version information"

private const val INDEXES_CMD_NAME = "indexes"
private const val INDEXES_CMD_DESCRIPTION = "Generates project shared indexes and makes them ready to be uploaded"

private const val SERVER_CMD_NAME = "server"
private const val SERVER_CMD_DESCRIPTION = "Starts server on local indexes"

private const val BOOST_CMD_NAME = "boost"
private const val BOOST_CMD_DESCRIPTION = "Measures indexing speed up with project shared indexes"

private const val IJ_OPT_NAME = "--ij"
private const val IJ_OPT_DESCRIPTION = "Path to a locally installed IDE"

private const val PROJECT_OPT_NAME = "--project"
private const val PROJECT_OPT_DESCRIPTION = "Path to a project"

private const val PORT_OPT_NAME = "--port"
private const val PORT_OPT_DESCRIPTION = "Port to run local server"

private const val DATA_DIR_OPT_NAME = "--data-directory"
private const val DATA_DIR_OPT_DESCRIPTION = "Directory to use for generation and server data"

private const val SERVER_DIR_OPT_NAME = "--server-directory"
private const val SERVER_DIR_OPT_DESCRIPTION = "Directory to use for server data"

private const val BASE_URL_OPT_NAME = "--base-url"
private const val BASE_URL_OPT_DESCRIPTION = "Indexes server URL is used in metadata until https://youtrack.jetbrains.com/issue/IDEA-269442 is fixed"

@Parameters(commandDescription = VERSION_CMD_DESCRIPTION)
private class VersionCommand {
  @Parameter(
    names = [IJ_OPT_NAME],
    description = IJ_OPT_DESCRIPTION,
    required = true,
    converter = ToPathConverter::class
  )
  var ij: Path? = null
}

@Parameters(commandDescription = INDEXES_CMD_DESCRIPTION)
private class IndexesCommand {
  @Parameter(
    names = [PROJECT_OPT_NAME],
    description = PROJECT_OPT_DESCRIPTION,
    required = true,
    converter = ToPathConverter::class,
    order = 0
  )
  var project: Path? = null

  @Parameter(
    names = [IJ_OPT_NAME],
    description = IJ_OPT_DESCRIPTION,
    required = true,
    converter = ToPathConverter::class,
    order = 1
  )
  var ij: Path? = null

  @Parameter(
    names = [BASE_URL_OPT_NAME],
    description = BASE_URL_OPT_DESCRIPTION,
    required = true,
    order = 2
  )
  var baseUrl: String? = null

  @Parameter(
    names = [DATA_DIR_OPT_NAME],
    description = DATA_DIR_OPT_DESCRIPTION,
    required = false,
    converter = ToPathConverter::class,
    order = 3
  )
  var dataDir: Path = Paths.get(DEFAULT_DATA_DIR)
}

@Parameters(commandDescription = SERVER_CMD_DESCRIPTION)
private class ServerCommand {
  @Parameter(names = [PORT_OPT_NAME], description = PORT_OPT_DESCRIPTION, required = false)
  var port: Int = DEFAULT_PORT

  @Parameter(
    names = [SERVER_DIR_OPT_NAME],
    description = SERVER_DIR_OPT_DESCRIPTION,
    required = false,
    converter = ToPathConverter::class
  )
  var serverDir: Path = Paths.get(DEFAULT_DATA_DIR).serverDir()
}

@Parameters(commandDescription = BOOST_CMD_DESCRIPTION)
private class BoostCommand {
  @Parameter(
    names = [PROJECT_OPT_NAME],
    description = PROJECT_OPT_DESCRIPTION,
    required = true,
    converter = ToPathConverter::class,
    order = 0
  )
  var project: Path? = null

  @Parameter(
    names = [IJ_OPT_NAME],
    description = IJ_OPT_DESCRIPTION,
    required = true,
    converter = ToPathConverter::class,
    order = 1
  )
  var ij: Path? = null

  @Parameter(
    names = [PORT_OPT_NAME],
    description = PORT_OPT_DESCRIPTION,
    required = false,
    order = 2
  )
  var port: Int = DEFAULT_PORT

  @Parameter(
    names = [DATA_DIR_OPT_NAME],
    description = DATA_DIR_OPT_DESCRIPTION,
    required = false,
    converter = ToPathConverter::class,
    order = 3
  )
  var dataDir: Path = Paths.get(DEFAULT_DATA_DIR)
}

private fun printHelp() {
  println(
    """
      Usage example: indexes-server --ij <IDE_DIR> --project <PROJECT_DIR>

      Commands:
        $BOOST_CMD_NAME
          $BOOST_CMD_DESCRIPTION
            Options:
            * $PROJECT_OPT_NAME
                $PROJECT_OPT_DESCRIPTION
            * $IJ_OPT_NAME
                $IJ_OPT_DESCRIPTION
              $PORT_OPT_NAME
                $PORT_OPT_DESCRIPTION
                Default: $DEFAULT_PORT
              $DATA_DIR_OPT_NAME
                $DATA_DIR_OPT_DESCRIPTION
                Default: ${Paths.get(DEFAULT_DATA_DIR).absolutePathString()}

        $INDEXES_CMD_NAME
          $INDEXES_CMD_DESCRIPTION
            Options:
            * $PROJECT_OPT_NAME
                $PROJECT_OPT_DESCRIPTION
            * $IJ_OPT_NAME
                $IJ_OPT_DESCRIPTION
            * $BASE_URL_OPT_NAME
                $BASE_URL_OPT_DESCRIPTION
              $DATA_DIR_OPT_NAME
                $DATA_DIR_OPT_DESCRIPTION
                Default: ${Paths.get(DEFAULT_DATA_DIR).absolutePathString()}

        $SERVER_CMD_NAME
          $SERVER_CMD_DESCRIPTION
            Options:
              $PORT_OPT_NAME
                $PORT_OPT_DESCRIPTION
                Default: $DEFAULT_PORT
              $SERVER_DIR_OPT_NAME
                $SERVER_DIR_OPT_DESCRIPTION
                Default: ${Paths.get(DEFAULT_DATA_DIR).serverDir().absolutePathString()}

        $VERSION_CMD_NAME
          $VERSION_CMD_DESCRIPTION
            Options:
            * $IJ_OPT_NAME
                $IJ_OPT_DESCRIPTION
    """.trimIndent()
  )
}

private class ToPathConverter : IStringConverter<Path> {
  override fun convert(value: String): Path = Paths.get(value)
}
